/*  This header file contains 3 member functions which generate random numbers each with a 
    different distribution.  This file is used by Generate.C
    
*/

#include <math.h>
#include <iostream.h>
#include "SpecialFunction.h"

class Generate{

public:

    Generate(): idum(-1),iy(0),sq(-1.0),alxm(-1.0),g(-1.0),oldm(-1.0)   {}

    float  Random();                  			// Returns a uniform random deviate between
     			          		// 0.0 and 1.0 (excluding the endpoints)

    float Gaussian(const float& mean,const float& sigma); // Returns a random Gaussian distribution
                                                                   	

    float Poisson(const float& mean);    		// Returns a random Poisson distribution

private:

    int iset;                   			// Data members needed for Gaussian
    double gset,fac,rsq,v1,v2;

    const int ia=16807;         			// Initializations needed for
    const long im=2147483647;   	 		// random number generator of
    const double am=(1.0/im);   			// Park and Miller
    const long iq=127773;
    const int ir=2836;
    const int ntab=32;
    const long ndiv=(1+(im-1)/ntab);
    const float eps=1.2e-7;
    const double rnmx=(1.0-eps);
    long idum;                      

    SpecFunc func;					//data members needed for Poisson
    int j;                   	
    long k;
    long iy;
    long iv[ntab];
    double temp;
    const double PI=3.141592654;
    double sq,alxm,g,oldm;
    double em,t,y;        
};

// The code to calculate these generators are in "Numerical Recipes in C The Art of Scientific Computing 
// Second Edition"
    
float Generate::Random() {				// Generates random numbers even distributed
						// between 0 and 1.  0 and 1 themself are not
   						// included.     
    if(idum <= 0 || !iy) {           			// In "Numerical Recipes" Ch 7-1 p.280
       	
        if (-(idum) < 1) idum=1;    			
        else idum= -(idum);           
        for (j=ntab+7;j>=0;j--) { 			  
            
            k=idum/iq;
            idum=ia*(idum-k*iq)-ir*k;
            if (idum < 0) idum +=im;
            if (j < ntab) iv[j]=idum;
        }
        iy=iv[0];
    }
    
    k=(idum)/iq;                      			
    idum=ia*(idum-k*iq)-ir*k;     			
    if (idum < 0) idum +=im;      			
    j=iy/ndiv;		     			
    iy=iv[j];                     			
    iv[j]=idum;                   			
    if ((temp=am*(iy)) > rnmx) return rnmx; 		 
    else return temp;                     		
                
}
  
float Generate::Gaussian(const float& mean,const float& sigma) {   // Generates random numbers with a
							// gaussian distribution.
							// In "Numerical Recipes" Ch 7-2 p.289

    if (idum < 0) iset=0;
    if(iset == 0) {
       do {
           v1=2.0*Random()-1.0;
           v2=2.0*Random()-1.0;
           rsq=v1*v1+v2*v2;        
           } while (rsq >=1.0 || rsq == 0.0);
           
       fac=sqrt(-2.0*sigma*sigma*log(rsq)/rsq);
       
       gset=v1*fac+mean;
       iset=1;
       return v2*fac+mean;
    } else { 
       iset=0;
       return gset;   
    }
}    

float Generate::Poisson(const float& mean) {		// Generates random numbers is a Poisson
						// distribution.
						// In "Numerical Recipes" Ch 7-3 p.294              
    if (mean < 12.0) {
       
       if (mean != oldm) {
          
           oldm=mean;
           g=exp(-mean);
       }
       em= -1;
       t=1.0;
       
       do {
           
           ++em;
           t *= Random();
           
       } while (t > g);
 
    } else {
    
       if (mean != oldm) {
          
          oldm=mean;
          sq=sqrt(2.0*mean);
          alxm=log(mean);
          g=mean*alxm - func.Gammln(mean+1.0);
          
       }
       
       do {
        
          do {
            
              y=tan(PI*Random());
              em=sq*y + mean;
              
           }  while (em < 0.0);
           
           em=floor(em);
           t=0.9*(1.0 + y*y)*exp(em*alxm - func.Gammln(em + 1.0) -g);
           
       } while (Random() > t);
    }             
        
    return em;          
}

